﻿/*
Copyright (c) 2010 Lance Ward
This software is licensed under the MIT license, except where otherwise noted.
For complete license details, refer to license.txt
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using ADNLib;

namespace ADNCanvas
{
    public enum AnimPointEditTarget
    {
        PathMarker,
        PathControlPoint,
        PathSegment,
        ObjectControlPoint,
        Object,
        None
    }

    public class ToolAnimPointEdit : Tool
    {
        private PointEditTarget target = PointEditTarget.None;
		private List<PathPoint> selPathMarkers = new List<PathPoint>();
		private PointF selObjControlPoint = PointF.Empty;
		private PointF selPathControlPoint = PointF.Empty;
		private int selPathSegment = -1;

        public ToolAnimPointEdit() 
        { 
            toolType = DrawingToolType.AnimPointEdit; 
        }

        //only allow one object
        public override List<XObject> SelectedObjects
        {
            get { return base.SelectedObjects; }
            set 
            { 
                base.SelectedObjects = value;
                if (selObjects.Count > 0)
                {
                    XObject o = selObjects[0];
                    selObjects.Clear();
                    selObjects.Add(o);
                }
            }
        }

		public List<PathPoint> SelectedPathMarkers
		{
			get { return selPathMarkers; }
		}

		public override RectangleF EraseRect
		{
			get { return GetEraseRect(); }
		}

        //this is a duplicate of XGroup.Bounds
		private RectangleF GetEraseRect()
		{
			RectangleF rect;

			if (selObjects != null && selObjects.Count > 0 && mode != ToolMode.None)
			{
				Shape firstShape = (Shape)selObjects[0];
				rect = firstShape.Bounds;

				for (int i = 1; i < selObjects.Count; i++)
				{
					Shape s = (Shape)selObjects[i];
					rect = RectangleF.Union(rect, s.Bounds);
				}

				RectangleF prevRect = new RectangleF(prevPoint.X, prevPoint.Y, 0, 0);
                rect = RectangleF.Union(rect, prevRect);

				rect.Inflate(20, 20);
			}
			else if (currentShape != null) { rect = currentShape.Bounds; }
			else { rect = new RectangleF(); }

			return rect;
		}

        public override void MouseDown(PointF location, System.Windows.Forms.MouseButtons button, int clicks)
        {
            int index;
            if (Control.ModifierKeys == Keys.Control)
            {
                index = HitTestObjects(location);
                if (index > -1)
                {
                    selObjects.Clear();
                    selObjects.Add(xObjects[index]);
                    OnSelectionChanged();
                }
            }
            else if (selObjects != null && selObjects.Count > 0)
            {
				Shape s = (Shape)selObjects[0];
                //hit test path markers
                index = HitTestMarkers(location);
				if (index >= 0) 
				{
					ClearSelections();
					selPathMarkers.Add(s.Path.Points[index]);
					prevPoint = location;
					target = PointEditTarget.PathMarker;
					return;
				}

                //hit test object control point
				index = HitTestObjectControlPoint(location);
				if (index >= 0) 
				{ 
					target = PointEditTarget.ObjectControlPoint; 
				}

                //hit test path control points
                index = HitTestPathControlPoints(location);
				if (index >= 0) 
				{ 
					target = PointEditTarget.BezierControlPoint; 
				}

				//hit test the lines
				index = HitTestPathSegments(location);
				if (index >= 0) 
				{ 
					target = PointEditTarget.PathSegment; 
				}
            }

        }

		private void ClearSelections()
		{
			selObjControlPoint = PointF.Empty;
			selPathControlPoint = PointF.Empty;
			selPathSegment = -1;
			selPathMarkers.Clear();
		}

        public override void MouseMove(PointF location, System.Windows.Forms.MouseButtons button, int clicks)
        {
			switch (target)
			{
				case PointEditTarget.PathMarker:
					{
						Shape s = (Shape)selObjects[0];
						PointF moveAmount = new PointF(location.X - prevPoint.X, location.Y - prevPoint.Y);
						s.MoveMarker(moveAmount, selPathMarkers[0]);
						prevPoint = location;
						OnRefresh(true);
					}
					break;
				case PointEditTarget.PathSegment:
					{

					}
					break;
				case PointEditTarget.BezierControlPoint:
					{
					}
					break;
				case PointEditTarget.ObjectControlPoint:
					{

					}
					break;
			}
        }

        public override void MouseUp(PointF location, System.Windows.Forms.MouseButtons button, int clicks)
        {
			target = PointEditTarget.None;
        }

        protected override void OnCursorChange(Cursor cursor)
        {
            base.OnCursorChange(cursor);
        }

        protected override void OnRefresh(bool forcePaint)
        {
            base.OnRefresh(forcePaint);
        }

        //TODO: this needs to be moved to Tool, along with the XObjects list
        private void SelectObjects(RectangleF r)
        {
            if (xObjects != null)
            {
                for (int i = xObjects.Count - 1; i >= 0; i--)
                {
                    if (xObjects[i].HitTest(parentSurface, r))
                    {
                        selObjects.Add(xObjects[i]);
                    }
                }
            }
        }

        private int HitTestMarkers(PointF p)
        {
            for (int i = 0; i < selObjects.Count; i++)
            {
                int index = selObjects[i].HitTestPathMarkers(parentSurface, p);
                if (index >= 0) { return index; }
            }

            return -1;
        }

        private int HitTestObjectControlPoint(PointF p)
        {
            for (int i = 0; i < selObjects.Count; i++)
            {
                int index = selObjects[i].HitTestPathMarkers(parentSurface, p);
                if (index >= 0) { return index; }
            }

            return -1;
        }

        private int HitTestPathControlPoints(PointF p)
        {
            for (int i = 0; i < selObjects.Count; i++)
            {
                int index = selObjects[i].HitTestPathMarkers(parentSurface, p);
                if (index >= 0) { return index; }
            }

            return -1;
        }

		private int HitTestPathSegments(PointF p)
		{
			for (int i = 0; i < selObjects.Count; i++)
			{
				int index = selObjects[i].HitTestPathMarkers(parentSurface, p);
				if (index >= 0) { return index; }
			}

			return -1;
		}
    }
}
